// ------------------------------------------------------
// RC - Build an object from 'row' and 'column' 
// ------------------------------------------------------

function RC(row, column)
{
   return({ r: row, c: column });
}

// ------------------------------------------------------
// idToRC - Find the row and column for 'theID'
// ------------------------------------------------------

function idToRC(theID)
{
   var result = null;
   var match  = /^b_(\d+)_(\d+)$/.exec(theID);

   if(match)
      result = RC(match[1], match[2]);

   return(result);
}

// ------------------------------------------------------
// gid - Return the element with 'id'
// ------------------------------------------------------

function gid(id)
{
   return(document.getElementById(id));
}

// ------------------------------------------------------
// setRC - Set the label of the button at (row, column)
//         to 'value'
// ------------------------------------------------------

function setRC(row, column, value)
{
   var id = "b" + "_" + row + "_" + column;

   gid(id).value = value;
}

// ------------------------------------------------------
// getBoardSize - Determine the size of our tic tac toe
//                board by seeking the largest row or
//                column value in any button element id
// ------------------------------------------------------

function getBoardSize()
{
   var form = gid("the-form");
   var size = 0;

   for(var i = 0; i < form.elements.length; i ++)
   {
      var rc = idToRC(form.elements[i].id);

      if(rc != null)
         size = Math.max(size, Math.max(rc.r, rc.c));
   }

   return(size + 1);
}

// ------------------------------------------------------
// isVictory - Return true if player 'current' has won
// ------------------------------------------------------

function isVictory(current, board)
{
   // check horizontal/vertical

   var size = board.length;

   for(var i = 0; i < size; i ++)
   {
      var c1 = 0;
      var c2 = 0;

      for(var j = 0; j < size; j ++)
      {
         if(board[i][j] == current)
            c1++;

         if(board[j][i] == current)
            c2++;
      }

      if(c1 == size || c2 == size)
         return(true);
   }

   // check, diagonal

   var c1 = 0;
   var c2 = 0;

   for(var i = 0, j = size - 1; i < size; i ++, j --)
   {
      if(board[i][i] == current)
         ++c1;

      if(board[i][j] == current)
         ++c2;
   }

   return(c1 == size || c2 == size);
}

// ------------------------------------------------------
// isFull - Return true if 'board' is entirely filled
// ------------------------------------------------------

function isFull(board)
{
   var size = board.length;

   for(var i = 0; i < size; i ++)
      for(j = 0; j < size; j ++)
         if(board[i][j] == "")
            return(false);

   return(true);
}

// ------------------------------------------------------
// getOther - Given 'current', return back the symbol
//            for the opposing player
// ------------------------------------------------------

function getOther(current)
{
   return(current == 'X' ? 'O' : 'X');
}

// ------------------------------------------------------
// doHuman - Choose 'theButton' and set the label to
//           'mCurrent'.
// ------------------------------------------------------

function doHuman(theButton)
{
   var rc = idToRC(theButton.id);

   if(rc != null)
      gGame.select(rc.r, rc.c);
}

// ------------------------------------------------------
// doComputer - Have the computer decide which move to
//              make and then show that move
// ------------------------------------------------------

function doComputer()
{
   var location = gGame.getComputerMove();

   gGame.select(location.r, location.c);
}

// ------------------------------------------------------
// makeGame - Create our game object
// ------------------------------------------------------

function makeGame()
{
   // Our game object's private variables

   var mCurrent  = 'X';
   var mGameOver = false;
   var mSize     = getBoardSize();
   var mBoard    = [];

   for(var i = 0; i < mSize; i ++)
      mBoard.push(new Array(mSize).fill(""));

   // Create the game object and then return it

   var theObject =
   {
      apply: function()
      {
         for(var i = 0; i < mSize; i ++)
            for(var j = 0; j < mSize; j ++)
               setRC(i, j, mBoard[i][j]);
      },

      reset: function()
      {
         for(var i = 0; i < mSize; i ++)
            for(var j = 0; j < mSize; j ++)
               mBoard[i][j] = "";

         mCurrent = 'X';
         this.apply();
         mGameOver = false;
      },

      copyBoard: function()
      {
         return(JSON.parse(JSON.stringify(mBoard)));
      },

      select: function(row, column)
      {
         if(mGameOver)
         {
            alert("Game Over - Press 'New Game' To Reset");
         }
         else if(mBoard[row][column] != "")
         {
            alert("Location Already Selected");
         }
         else
         {
            mBoard[row][column] = mCurrent;
            this.apply();

            if(isVictory(mCurrent, mBoard))
            {
               alert("Winner: " + mCurrent);
               mGameOver = true;
            }
            else if(isFull(mBoard))
            {
               alert("Tie");
               mGameOver = true;
            }
            else
            {
               mCurrent = getOther(mCurrent);
            }
         }
      },

      getComputerMove: function()
      {
         var board = this.copyBoard();
         var size  = board.length;

         // Can 'mCurrent' win with the next move?

         for(var i = 0; i < size; i ++)
         {
            for(var j = 0; j < size; j ++)
            {
               if(board[i][j] == "")
               {
                  board[i][j] = mCurrent;

                  if(isVictory(mCurrent, board))
                     return(RC(i, j));

                  board[i][j] = "";
               }
            }
         }

         // Can 'other' win with the next move?

         var other = getOther(mCurrent);

         for(var i = 0; i < size; i ++)
         {
            for(var j = 0; j < size; j ++)
            {
               if(board[i][j] == "")
               {
                  board[i][j] = other;

                  if(isVictory(other, board))
                     return(RC(i, j));

                  board[i][j] = "";
               }
            }
         }

         // center square (if our size is odd), then corners

         var list = [];

         if(size % 2 != 0)
         {
            var index = Math.floor(size / 2);

            list.push(RC(index, index));
         }

         list.push(RC(0,        0));
         list.push(RC(0,        size - 1));
         list.push(RC(size - 1, size - 1));
         list.push(RC(size - 1, 0));

         for(var i = 0; i < list.length; i ++)
         {
            var r = list[i].r;
            var c = list[i].c;

            if(board[r][c] == "")
               return(RC(r, c));
         }

         // Just choose an open spot at random

         list = [];

         for(var i = 0; i < size; i ++)
            for(var j = 0; j < size; j ++)
               if(board[i][j] == "")
                  list.push(RC(i, j));

         // If no open locations, use (0, 0)

         var rc = null;

         if(list.length == 0)
            rc = RC(0, 0);

         else
            rc = list[Math.floor(Math.random() * list.length)];

         return(rc);
      }
   };

   return(theObject);
}

// ----------------------------
// Script Begins Here
// ----------------------------

var gGame = makeGame();

